
Require Import Terms.
Require Import LNaVSyntax.

(** * Big-Step Semantics for LambdaNaV *)

(** Helper functions that help us reduce the number of rules in the
    big-step semantics. This is important since the number of subcases
    in our non-interference proof is quadratic in the number of
    overlapping rules (i.e. not fully syntax-directed). *)

Definition empty := (V vEmpty,bot).
Definition push loc pc s := (V (vPush loc pc s),bot). (* bot, or the pc? *)

Definition propagate_nav (b : Box) (s : Atom): Box :=
  match b with
  | V _ => D eType s empty
  | D e oldS pt => D e oldS pt
  end.  

Definition bop_lb_box f (b' b'' : Box) (s : Atom) : Box :=
  match b', b'' with
  | (V (VConst (CLab l'))), (V (VConst (CLab l''))) => f l' l''
  | V (VConst (CLab l')), V _ => D eType s empty
  | V (VConst (CLab l')), D e oldS pt => D e oldS pt
  | V _, _ => D eType s empty
  | D e oldS pt, _ => D e oldS pt
  end.

Definition eq_box (b' b'' : Box) (s:Atom) : Box :=
  match b', b'' with
  | (V (VConst c')), (V (VConst c'')) =>
    if beq_const c' c'' then V vTrue else V vFalse
  | V (VConst v'), V _ => D eType s empty
  | V (VConst v'), D e oldS pt => D e oldS pt
  | V _, _ => D eType s empty
  | D e oldS pt, _ => D e oldS pt
  end.

Definition bop_box bo :=
  match bo with
  | BEq => eq_box
  | BCmp => bop_lb_box (fun l' l'' => if flows_dec l' l'' then V vTrue else V vFalse)
  | BJoin => bop_lb_box (fun l' l'' => V (VConst (CLab (join l' l''))))
  end.

Definition mk_nav_box (b : Box) (s:Atom) : Box :=
  match b with
  | V (VConst (CExcp e)) => D e s empty
  | _ => propagate_nav b s
  end.

Definition to_sum_box (b : Box) : Box :=
  match b with
  | V v => V (vInl (V v, bot))
  | D e st pt => V (vInr (V (VProd (V (vExcp e),bot) (V (VProd st pt),bot)), bot))
  end.

Definition tag_box (b : Box) : Box :=
  match b with
  | V v => V (VConst (CTag (tag_of v)))
  | D e st pt => D e st pt
  end.

Definition trace_propagation b loc (pc : Lab) :=
  match b with
    | V _ => b
    | D e st pt => D e st (push loc bot pt)
  end.

(** The evaluation judgment. *)

Reserved Notation "r ; s |= t , pc1 ==> a , pc2" (at level 80).
Inductive eval : Env -> Atom -> Tm -> Lab -> Atom -> Lab -> Prop :=
  | eval_var : forall r s x a pc,
      maps r x a -> 
      r; s |= (TVar x),pc ==> a, pc
  | eval_const : forall r s c pc,
      r; s |= TConst c, pc ==> (V (VConst c), bot), pc
  | eval_let : forall r s x t t' pc pc' pc'' a a',
      r; s |= t, pc ==> a, pc' ->
      (x,a)::r; s |= t', pc' ==> a', pc'' ->
      r; s |= TLet x t t', pc ==> a', pc''
  | eval_abs : forall r s x t pc,
      r;s |= (TAbs x t), pc ==> (V (VClos r x t), bot), pc
  | eval_app : forall r s x' x'' pc a' pc' r' x t a l,
      maps r x' (V (VClos r' x t), l) ->
      maps r x'' a ->
      (x,a)::r'; s |= t, (pc \_/ l) ==> a', pc' ->
      r; s |= TApp x' x'', pc ==> a', pc'
  (* optimized version (mixes 2 rules in 1) *)
  | eval_app_no_abs : forall r s x' x'' pc b l,
      maps r x' (b, l) ->
      ~(is_box is_abs b) ->
      r; s |= TApp x' x'', pc ==> (propagate_nav b s, bot), (pc \_/ l) (* trace *)
  | eval_prod : forall r s x' x'' pc a' a'',
      maps r x' a' ->
      maps r x'' a'' ->
      r; s |= TProd x' x'', pc ==> (V (VProd a' a''),bot), pc
  | eval_proj : forall r s d x pc a' a'' l,
      maps r x (V (VProd a' a''),l) ->
      r; s |= TProj d x, pc ==> d_choose d a' a'', (pc \_/ l)
  | eval_proj_no_prod : forall r s d x pc b l,
      maps r x (b,l) ->
      ~(is_box is_prod b) ->
      r; s |= TProj d x, pc ==> (propagate_nav b s, bot), (pc \_/ l)
  | eval_inx : forall r s d x pc a,
      maps r x a ->
      r; s |= TInx d x, pc ==> (V (VInx d a),bot), pc
  | eval_match : forall r s x x' t' t'' a' pc pc' a l d,
      maps r x (V (VInx d a),l) ->
      (x',a)::r; s |= d_choose d t' t'', pc\_/l ==> a', pc' ->
      r; s |= TMatch x x' t' t'', pc ==> a', pc'
  | eval_match_no_sum : forall r s x x' t' t'' pc l b,
      maps r x (b,l) ->
      ~(is_box is_sum b) ->
      r; s |= TMatch x x' t' t'', pc ==> (propagate_nav b s, bot), (pc\_/l) (* trace *)
  | eval_tag : forall r s x pc b l,
      maps r x (b, l) ->
      r; s |= TTag x, pc ==> (tag_box b, l), pc
  (* optimized version (mixes 3 x 5 rules in 1) *)
  | eval_bop : forall r s bo x' x'' pc b' b'' l' l'',
      maps r x' (b', l') ->
      maps r x'' (b'', l'') ->
      r; s |= (TBOp bo x' x''), pc ==> (bop_box bo b' b'' s, l' \_/ l''), pc (* trace *)
  (* optimized version (mixes 2 rules in 1) *)
  | eval_bracket : forall r s x t pc pc' b l l' l'',
      maps r x (V (VConst (CLab l)), l') ->
      r; s |= t, (pc\_/l') ==> (b, l''), pc' ->
      r; s |= TBracket x t, pc ==> (* trace *)
              (if flows_dec (l'' \_/ pc') (l \_/ (pc\_/l')) then b else D eBracket s empty, l), (pc\_/l')
  (* optimized version (mixes 2 rules in 1) *)
  | eval_bracket_no_lab : forall r s x t pc b l',
      maps r x (b, l') ->
      ~(is_box is_lab b) ->
      r; s |= TBracket x t, pc ==> (propagate_nav b s,bot), (pc\_/l') (* trace *)
  | eval_label_of : forall r s x b l pc,
      maps r x (b, l) ->
      r; s |= (TLabelOf x), pc ==> (V (VConst (CLab l)), bot), pc
  | eval_get_pc : forall r s pc,
      r; s |= TGetPc, pc ==> (V (VConst (CLab pc)), bot), pc
  (* optimized version (mixes 3 rules in 1) *)
  | eval_mk_nav : forall r s x pc b l,
      maps r x (b, l) ->
      r; s |= TMkNav x, pc ==> (mk_nav_box b s, l), pc (* trace *)
  (* optimized version (mixes 2 rules in 1) *)
  | eval_to_sum : forall r s x pc b l,
      maps r x (b,l) ->
      r; s |= TToSum x, pc ==> (to_sum_box b, l), pc
  (* stack trace operations *)
  | eval_loc : forall r s loc t pc b l pc',
      r; push loc bot s |= t, pc ==> (b,l), pc' ->
      r; s |= TLoc loc t, pc ==> (trace_propagation b loc bot,l), pc'
where "r ; s |= t , pc1 ==> a , pc2" := (eval r s t pc1 a pc2).
Hint Constructors eval.

Tactic Notation "eval_cases" tactic(first) ident(c) :=
  first; [
  (* perl -ne 'print "    Case_aux c \"$1\" |\n" if /\| (eval_\w+)\s/' < LNaVBigStep.v *)
    Case_aux c "eval_var" |
    Case_aux c "eval_const" |
    Case_aux c "eval_let" |
    Case_aux c "eval_abs" |
    Case_aux c "eval_app" |
    Case_aux c "eval_app_no_abs" |
    Case_aux c "eval_prod" |
    Case_aux c "eval_proj" |
    Case_aux c "eval_proj_no_prod" |
    Case_aux c "eval_inx" |
    Case_aux c "eval_match" |
    Case_aux c "eval_match_no_sum" |
    Case_aux c "eval_tag" |
    Case_aux c "eval_bop" |
    Case_aux c "eval_bracket" |
    Case_aux c "eval_bracket_no_lab" |
    Case_aux c "eval_label_of" |
    Case_aux c "eval_get_pc" |
    Case_aux c "eval_mk_nav" |
    Case_aux c "eval_to_sum" |
    Case_aux c "eval_loc"
  ].
