Require Import Terms.
Require Import LNaVSyntax.
Require Import LNaVBigStep.

(** * Equivalences *)

(** Low-equivalence judgments. *)
Inductive eq_atom : Atom -> Atom -> Lab -> Prop :=
| eq_a : forall b1 l1 b2 l2 l,
  l1 = l2 ->
  ((l1 <: l \/ l2 <: l) -> eq_box b1 b2 l) ->
  eq_atom (b1,l1) (b2,l2) l
with eq_box : Box -> Box -> Lab -> Prop :=
| eq_v : forall v1 v2 l, eq_val v1 v2 l -> eq_box (V v1) (V v2) l
| eq_d : forall e s1 s2 p1 p2 l, 
  eq_atom s1 s2 l -> eq_atom p1 p2 l -> eq_box (D e s1 p1) (D e s2 p2) l
with eq_val : Val -> Val -> Lab -> Prop :=
| eq_vconst : forall c' l, eq_val (VConst c') (VConst c') l
| eq_vprod : forall afst asnd afst' asnd' l,
  eq_atom afst afst' l -> eq_atom asnd asnd' l -> 
  eq_val (VProd afst asnd) (VProd afst' asnd') l
| eq_vinx : forall d a a' l, 
  eq_atom a a' l -> 
  eq_val (VInx d a) (VInx d a') l
| eq_vclos : forall r1 r2 x t l,
  eq_env r1 r2 l ->
  eq_val (VClos r1 x t) (VClos r2 x t) l
with eq_env : Env -> Env -> Lab -> Prop :=
| eq_e_nil : forall l, eq_env nil nil l
| eq_e_cons : forall x a1 a2 r1 r2 l,
  eq_atom a1 a2 l -> eq_env r1 r2 l -> eq_env ((x,a1)::r1) ((x,a2)::r2) l.
Hint Constructors eq_atom eq_box eq_val eq_env.
Scheme eq_atom_ind' := Minimality for eq_atom Sort Prop
with eq_box_ind' := Minimality for eq_box Sort Prop
with eq_val_ind' := Minimality for eq_val Sort Prop
with eq_env_ind' := Minimality for eq_env Sort Prop.

Combined Scheme eq_mutind from eq_val_ind', eq_box_ind', eq_atom_ind', eq_env_ind'.

(** Low-equivalence judgments, that take the pc label into account. *)
Definition eq_atom' a1 pc1 a2 pc2 l :=
  (pc1 <: l \/ pc2 <: l) ->
  (pc1 = pc2 /\ eq_atom a1 a2 l).
Definition eq_env' r1 pc1 r2 pc2 l :=
  (pc1 <: l \/ pc2 <: l) ->
  (pc1 = pc2 /\ eq_env r1 r2 l).

(** * Preliminary lemmas. *)

Lemma maps_eq_env : forall x a1 a2 r1 r2 l,
  eq_env r1 r2 l ->
  maps r1 x a1 ->
  maps r2 x a2 ->
  eq_atom a1 a2 l.
Proof.
  intros x a1 a2 r1 r2 l Heq_env Hmaps1 Hmaps2.
  induction Heq_env. invs Hmaps1.
  invs Hmaps1; invs Hmaps2. destruct (beq_nat x x0).
  congruence. auto.
Qed.

Lemma maps_eq_env' : forall x a1 a2 r1 r2 pc1 pc2 l,
  eq_env' r1 pc1 r2 pc2 l ->
  maps r1 x a1 ->
  maps r2 x a2 ->
  eq_atom' a1 pc1 a2 pc2 l.
Proof.
  intros x a1 a2 r1 r2 pc1 pc2 l Heq_env' Hmaps1 Hmaps2 Hflows.
  apply Heq_env' in Hflows. intuition eauto using maps_eq_env.
Qed.

Lemma eq_env'_cons_inv_env : forall x a1 a2 r1 r2 pc1 pc2 l,
  eq_env' ((x,a1) :: r1) pc1 ((x,a2) :: r2) pc2 l ->
  eq_env' r1 pc1 r2 pc2 l.
Proof.
  intros x a1 a2 r1 r2 pc1 pc2 l H Hflows. apply H in Hflows as [? Henv].
  invsc Henv. auto.
Qed.

Lemma eq_env'_cons_inv_atom : forall x a1 a2 r1 r2 pc1 pc2 l,
  eq_env' ((x,a1) :: r1) pc1 ((x,a2) :: r2) pc2 l ->
  eq_atom' a1 pc1 a2 pc2 l.
Proof.
  intros x a1 a2 r1 r2 pc1 pc2 l H Hflows. apply H in Hflows as [? Henv].
  invsc Henv. auto.
Qed.

Lemma cons_eq_env' : forall x a1 a2 r1 r2 pc1 pc2 l,
  eq_atom' a1 pc1 a2 pc2 l ->
  eq_env' r1 pc1 r2 pc2 l ->
  eq_env' ((x,a1) :: r1) pc1 ((x,a2) :: r2) pc2 l.
Proof.
  intros x a1 a2 r1 r2 pc1 pc2 l Hatom Henv Hflows.
  specialize (Hatom Hflows). specialize (Henv Hflows).
  intuition auto.
Qed.

(** Reflexivity of low-equivalence. Later, we prove this is actually an
    equivalence relation. *)

Lemma eq_refl :
  (forall v l, eq_val v v l) /\
  (forall b l, eq_box b b l) /\
  (forall a l, eq_atom a a l) /\
  (forall r l, eq_env r r l).
Proof.
 apply val_box_atom_env_mutind; eauto.
Qed.

Lemma eq_val_refl : forall v l, eq_val v v l.
Proof. pose proof eq_refl. intuition. Qed.

Lemma eq_box_refl : forall b l, eq_box b b l.
Proof. pose proof eq_refl. intuition. Qed.

Lemma eq_atom_refl : forall a l, eq_atom a a l.
Proof. pose proof eq_refl. intuition. Qed.

Lemma eq_env_refl : forall r l, eq_env r r l.
Proof. pose proof eq_refl. intuition. Qed.

Lemma intro_and2 : forall (P P1 P2: Prop),
  (P -> (P1 /\ P2)) ->
  (P -> P1) /\ (P -> P2).
Proof. tauto. Qed.

Lemma eq_val_eq_tag : forall v1 v2 l,
  eq_val v1 v2 l ->
  tag_of v1 = tag_of v2.
Proof.
  intros v1 v2 l Heq.
  inversion Heq; eauto.
Qed.

Lemma eq_box_eq_tag : forall b1 b2 l,
  eq_box b1 b2 l ->
  eq_box (tag_box b1) (tag_box b2) l.
Proof. intros. invs H; unfold tag_box; auto. rewrite (eq_val_eq_tag v1 v2 l H0); auto. Qed.

Lemma or_introrefl : forall (P : Prop),
  P -> P \/ P.
Proof. left. assumption. Qed.

Lemma or_elimrefl : forall (P : Prop),
  P \/ P -> P.
Proof. intros P H. destruct H; assumption. Qed.

(** Monotonicity of the pc label. This property is essential for the
    non-interference proof to go through. *)
Lemma pc_eval_monotonic : forall r s t pc a pc',
  r; s |= t, pc ==> a, pc' -> pc <: pc'.
Proof.
  intros r s t pc a pc' Heval.
  (eval_cases (induction Heval) Case);
  eauto 4 using flows_refl, flows_trans, join_1_rev.
Qed.

(* Binary operations respect equality *)
Lemma eq_box_eq_bop : forall b b11 b12 b21 b22 s1 s2 l,
  eq_box b11 b21 l ->
  eq_box b12 b22 l ->
  eq_atom s1 s2 l ->
  eq_box (bop_box b b11 b12 s1) (bop_box b b21 b22 s2) l.
Proof.
  intros b b11 b12 b21 b22 a1 a2 l Heq1 Heq2 Heqa.
  invsc Heq1; eauto; invsc Heq2; eauto;
  destruct b; invsc H; invsc H0; simpl;
    try destruct c'; try destruct c'0; eauto using eq_box_refl, eq_atom_refl.
Qed.

Lemma eq_box_eq_propagate : forall b1 b2 s l,
  eq_box b1 b2 l ->
  eq_box (propagate_nav b1 s) (propagate_nav b2 s) l.
Proof. intros b1 b2 s l Heq. invsc Heq; simpl; eauto using eq_box_refl. Qed.

(** * Non-interference *)

Ltac cleanup :=
  repeat match goal with
           | H : ?x = ?x |- _ => clear H
           | H1 : ?l1 <: ?l2, H2 : ?l1 <: ?l2 |- _ => clear H2
         end.

Ltac open_bounds := 
  repeat match goal with 
           | H : (?l1 <: ?l \/ ?l1 <: ?l) |- _ => apply or_elimrefl in H
           | H : (?l1 <: ?l \/ ?l2 <: ?l) |- _ => destruct H
         end.

Ltac notHyp P :=
  match goal with
    | [ _ : P |- _ ] => fail 1
    | _ =>
      match P with
        | ?P1 /\ ?P2 => first [ notHyp P1 | notHyp P2 | fail 2 ]
        | ?P1 \/ ?P2 => first [ notHyp P1; notHyp P2 | fail 2 ]
        | _ => idtac
      end
  end.

Ltac pc_monotonicity :=
  repeat match goal with
           | H : ?r; _ |= ?t,?pc ==> _,?pc' |- _ => 
             notHyp (pc <: pc'); pose proof (pc_eval_monotonic _ _ _ _ _ _ H)
         end.

Ltac solve_bounds :=
  pc_monotonicity; open_bounds; solve 
    [eauto using flows_trans |
     eauto using join_1_rev, join_2_rev |
     eauto using flows_trans, join_1_rev, join_2_rev |
     eauto using flows_trans, join_1_rev, join_2_rev, join_minimal].

Ltac unprime :=
  unfold eq_atom',eq_env' in *;
  repeat match goal with
           | H : (?l1 <: ?l \/ ?l2 <: ?l)->?P |- _ =>
             let Hflow := fresh "Hflow" in
             assert (l1 <: l \/ l2 <: l) as Hflow by solve_bounds;
             specialize (H Hflow); clear Hflow;
             let Heqpc := fresh "Heqpc" in
             try (destruct H as [Heqpc H]; invsc Heqpc)
         end;
  cleanup.

Ltac solve_vars :=
  repeat match goal with
           | Henv : eq_env ?r1 ?r2 ?l, Hmap1 : maps ?r1 ?x ?a1, Hmap2 : maps ?r2 ?x ?a2 |- _ =>
             notHyp (eq_atom a1 a2 l);
             let Heqa := fresh "Heq_a" in
             pose proof (maps_eq_env _ _ _ _ _ _ Henv Hmap1 Hmap2) as Heqa; eauto
         end; cleanup.

Ltac same_atoms :=
  unprime;
  repeat match goal with | Heqa : eq_atom ?a1 ?a2 ?l |- _ => invsc Heqa end;
  unprime.

Ltac same_values := 
  repeat match goal with | Heqa : eq_box ?b1 ?b2 ?l |- _ => invsc Heqa end;
  repeat match goal with | Heqa : eq_val ?v1 ?v2 ?l |- _ => invsc Heqa end.

Ltac unbox_vars := solve_vars; same_atoms.

Ltac equal_vars := unbox_vars; same_values.

Ltac propagate :=
  match goal with
    | H : eq_box ?b1 ?b2 ?l |- context[propagate_nav ?b1 ?s] =>
      apply (eq_box_eq_propagate b1 b2 s) in H; eauto 7 using eq_atom_refl
  end.

Ltac useIH userL := 
  match goal with
    | IH : (forall r2 pc2 res2 pc2' l,
            r2; _ |= ?t, pc2 ==> res2, pc2' ->
            (?pc1 <: l \/ pc2 <: l -> ?pc1 = pc2 /\ eq_env ?r1 r2 l) ->
            ?result) ,
      Heval1 : ?r1; _ |= ?t, ?pc1 ==> ?res1, ?pc1', 
      Heval2 : _; _ |= ?t, ?pc1 ==> _, _ |- _ => 
        let Henv := fresh "Heqenv" in
        let Ha := fresh "Heqa" in
        let Htrace := fresh "Heqs" in
        apply (IH _ _ _ _ userL) in Heval2 as [Henv Ha]; clear IH;
        try solve [intro; eauto]
  end; unprime.

Ltac withPc := 
  let Hpc := fresh "Hpc" in
  apply intro_and2; intro Hpc.

Ltac spurious :=
  try contradiction;
  match goal with
    | H : ~ is_box _ _ |- _ =>
      simpl in H; exfalso; auto
    | H : eq_box (D _ _) (V _) _ |- _ => invsc H
    | H : eq_box (V _) (D _ _) _ |- _ => invsc H
  end.

Ltac NICase name :=
  Case name; withPc; 
  match goal with | H : ?r2; _ |= _,_ ==> _,_ |- context[eq_env _ ?r2 _] => invsc H end; 
  unprime.

Ltac focus_atom := repeat (split; auto); intro; unprime.

(** Strengthened version of non-interference. *)
Lemma non_interference_strong :
  forall r1 s t pc1 a1 pc1',
    r1; s |= t, pc1 ==> a1, pc1' ->
      forall r2 pc2 a2 pc2' l,
        r2; s |= t, pc2 ==> a2, pc2' ->
          eq_env' r1 pc1 r2 pc2 l ->
          (eq_env' r1 pc1' r2 pc2' l /\
           eq_atom' a1 pc1' a2 pc2' l).
Proof.
intros r1 s t pc1 a1 pc1' H.
(eval_cases (induction H) Case);
intros r2 pc2 a2 pc2'; try (rename l into l1); intro l;
  intros Heval2 Heq_env.
NICase "eval_var". solve_vars. 
NICase "eval_const". auto.
NICase "eval_let". useIH l. useIH l. auto.
NICase "eval_abs". auto 6.
NICase "eval_app".
  SCase "success". equal_vars. useIH l. auto.
  SCase "type error". equal_vars. spurious.
NICase "eval_app_no_abs".
  SCase "success". equal_vars. spurious.
  SCase "type error". unbox_vars. propagate.
NICase "eval_prod". solve_vars. auto 6.
NICase "eval_proj".
  SCase "success". equal_vars. destruct d; auto.
  SCase "failure". equal_vars. spurious.
NICase "eval_proj_no_prod".
  SCase "success". equal_vars. spurious.
  SCase "failure". unbox_vars. propagate.
NICase "eval_inx". solve_vars. auto 6.
NICase "eval_match".
  SCase "success". equal_vars. useIH l. auto.
  SCase "failure". equal_vars. spurious.
NICase "eval_match_no_sum".
  SCase "success". equal_vars. spurious.
  SCase "failure". unbox_vars. propagate.
NICase "eval_tag". unbox_vars. focus_atom. apply eq_box_eq_tag; auto.
NICase "eval_bop". unbox_vars. focus_atom. apply eq_box_eq_bop; auto using eq_atom_refl. 
NICase "eval_bracket".
  SCase "success".
    pc_monotonicity. equal_vars. focus_atom. useIH l.
    remember (flows_dec (l'' \_/ pc') (l0 \_/ (pc2 \_/ l'0))) as f1.
    remember (flows_dec (l''0 \_/ pc'0) (l0 \_/ (pc2 \_/ l'0))) as f2.
    destruct f1; destruct f2; same_atoms; try solve [auto using eq_box_refl | spurious].
  SCase "type error". equal_vars. spurious.
NICase "eval_bracket_no_lab". 
  SCase "success". equal_vars. spurious.
  SCase "type error". unbox_vars. propagate.
NICase "eval_label_of". unbox_vars. auto.
NICase "eval_get_pc". auto.
NICase "eval_mk_nav". 
  unbox_vars. focus_atom. same_values; auto using eq_box_refl. constructor; auto.
NICase "eval_to_sum". 
  unbox_vars. focus_atom. same_values; unfold to_sum_box; repeat (constructor; auto).
NICase "eval_loc". 
  useIH l. unbox_vars. focus_atom. same_values; unfold trace_propagation, push; repeat (constructor; auto).
Qed.

(** Finally, non-interference. *)

Theorem non_interference: forall r1 s t pc1 a1 pc1' r2 pc2 a2 pc2' l,
    eq_env' r1 pc1 r2 pc2 l ->
    r1; s |= t, pc1 ==> a1, pc1' ->
    r2; s |= t, pc2 ==> a2, pc2' ->
    eq_atom' a1 pc1' a2 pc2' l.
Proof.
  intros r1 s t pc1 a1 pc1' r2 pc2 a2 pc2' l Heq_env Heval1 Heval2.
  pose proof (non_interference_strong _ _ _ _ _ _ Heval1  _ _ _ _ l Heval2 Heq_env)
    as [_ H].
  assumption.
Qed.
